#!/usr/bin/env python3.5

#####################################################################
## Author: Taka@FunIncomplete
## Updated: Apr-2020
## Visit: www.FunIncomplete.com
#####################################################################

### User configurations
Reset = False
endProgramAfter = 10

### Preparing LCD screen
import I2C_LCD_driver
lcd = I2C_LCD_driver.lcd()

### Importing other libraries
lcd.lcd_display_string('Importing Libs..'.ljust(16), 1, 0)
print('[INFO] Importing Libs...')

import cv2
import sqlite3
import RPi.GPIO as GPIO
from numpy import array as np_array
from subprocess import call
from os import getcwd
from os import remove
from time import sleep
from pathlib import Path
from datetime import date, datetime

call('sudo killall bluealsa', shell=True)
call('pulseaudio --start', shell=True)
call('pactl load-module module-bluetooth-discover', shell=True)
call('bluetoothctl << EOF\nconnect 41:42:87:63:FD:3E\nEOF', shell=True)

### Reset program
if Reset:
	lcd.lcd_display_string('Resetting Data'.ljust(16), 1, 0)
	if Path(ScriptLocation + '/face_recogniser.db').is_file():
		remove(ScriptLocation + '/face_recogniser.db')
	if Path(ScriptLocation + '/facetrainer.yml').is_file():
		remove(ScriptLocation + '/facetrainer.yml')
	if Path(ScriptLocation + '/voice.wav').is_file():
		remove(ScriptLocation + '/voice.wav')

### Initialising Variables
lcd.lcd_display_string('Init Variables'.ljust(16), 1, 0)
print('[INFO] Init Variables...') 

ScriptLocation = getcwd()
newTrainer = True
newDB = True
NameCursorBlink = True
endProgramCount = 0
mode = 'Detecting'
trainingImg = []
predictedIDset = set()
predictedID = -1
predictedName = '...'
name = ''
msg1 = ''
msg2 = ''
NewSpeech = ''
OldSpeech = ''
LCDLine1 = ''
LCDLine2 = ''
indicator = '>'
alphabet_index = 1
alphabet = ['','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z',' ','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z']
WeekDay = ["Mon","Tue","Wed","Thu","Fri","Sat","Sun"]

### Initialising buttons
lcd.lcd_display_string('Init Buttons...'.ljust(16), 1, 0)
print('[INFO] Init Buttons...') 
Button_1 = 17
Button_2 = 23
Button_3 = 24
Button_4 = 22
Button_5 = 27 

GPIO.setmode(GPIO.BCM)
GPIO.setup(Button_1, GPIO.IN, pull_up_down=GPIO.PUD_UP)
GPIO.setup(Button_2, GPIO.IN, pull_up_down=GPIO.PUD_UP)
GPIO.setup(Button_3, GPIO.IN, pull_up_down=GPIO.PUD_UP) 
GPIO.setup(Button_4, GPIO.IN, pull_up_down=GPIO.PUD_UP)
GPIO.setup(Button_5, GPIO.IN, pull_up_down=GPIO.PUD_UP) 

### Initialising face detector
lcd.lcd_display_string('Init Detector'.ljust(16), 1, 0)
print('[INFO] Init Detector...') 
face_cascade = cv2.CascadeClassifier(ScriptLocation + '/haarcascade_frontalface_default.xml')

### Initialising SQLite database
lcd.lcd_display_string('Preparing SQlite'.ljust(16), 1, 0)
print('[INFO] Preparing SQlite...')  

if Path(ScriptLocation + '/face_recogniser.db').is_file():
	newDB = False
	
conn = sqlite3.connect(ScriptLocation + '/face_recogniser.db')
c = conn.cursor()

if newDB:
	sql = '''CREATE TABLE 'people' ( 'name' TEXT, 'date_created' TEXT )'''
	c.execute(sql)

### Initialising face recogniser
lcd.lcd_display_string('Init Recogniser'.ljust(16), 1, 0)
print('[INFO] Init Recogniser...') 
recogniser = cv2.face.LBPHFaceRecognizer_create()
if Path(ScriptLocation + '/facetrainer.yml').is_file():
	recogniser.read(ScriptLocation + '/facetrainer.yml')
	newTrainer = False

################################### START Function Blocks ###################################

### Button 1
def button1_callback(channel):
	global cap
	global trainingImg
	global newTrainer
	global mode
	global name
	global indicator
	
	if mode in {'Detecting', 'Recording'}:
		if GPIO.input(Button_1):		
			if len(trainingImg) > 0:
				if newTrainer:
					mode = 'Name'
				else:
					mode = 'Predicting'
			else:
				mode = 'Detecting'
		else:	
			while(not GPIO.input(Button_1)):
				if '>'==indicator:
					indicator = '+'
				else:
					indicator = '>'	
				print('Recording')
				mode = 'Recording'
				for i in range(5):
					ret, frame = cap.read()
				frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
				faces = face_cascade.detectMultiScale(frame, 1.3, 5)
				for (x,y,w,h) in faces:
					trainingImg.append(frame[y:y+h, x:x+w])
					print('trainImg:', len(trainingImg))
					break
	elif mode in {'Name'}:
		print('Button 1 Name mode')
		if len(name)>0:
			mode = 'Saving'
		else:
			trainingImg = []
			mode = 'Detecting'

### Button 2
def button2_callback(channel):
	global mode
	global alphabet
	global alphabet_index
	
	if 'Name' == mode:
		if not GPIO.input(Button_2):
			print('Button 2 Name mode')
			alphabet_index-=1
			if alphabet_index  <  0:
				alphabet_index = len(alphabet) - 1
	elif mode in {'Predicting'}:
		mode = 'Updating'

### Button 3
def button3_callback(channel):
	global mode
	global alphabet
	global alphabet_index
	if 'Name' == mode:
		if not GPIO.input(Button_3):
			print('Button 3 Name mode')
			alphabet_index+=1
			if alphabet_index >= len(alphabet):
				alphabet_index = 0

### Button 4
def button4_callback(channel):  
	global mode
	global alphabet
	global name
	global alphabet_index
	if 'Name' == mode:
		if not GPIO.input(Button_4):
			print('Button 4 Name mode')
			if len(name)  <  11:
				name += alphabet[alphabet_index]
				alphabet_index = 0
	elif 'Predicting'== mode: #Pressed 'No' in predictedName mode
		print('Button 4 Predicting mode')
		if len(predictedIDset)==0:
			resetPrediction()
			mode = 'Name' #Name entry mode
		else:
			predictedIDset.discard(predictedID)
			if len(predictedIDset)==0:
				resetPrediction()
				mode = 'Name'
	
### Button 5
def button5_callback(channel):
	global endProgramCount
	global alphabet
	global name
	global alphabet_index
	global mode
	global trainingImg
	
	if 'Name' == mode:
		if not GPIO.input(Button_5):
			print('Button 5 Name mode')
			if len(name)>0:
				alphabet_index = alphabet.index(name[-1:])
				name = name[:-1]
			else:
				trainingImg = []
				mode = 'Detecting'

	else:
		if GPIO.input(Button_5):
			endProgramCount = 0
			mode = 'Detecting'
		else:
			while(not GPIO.input(Button_5)):
				sleep(1)
				endProgramCount += 1
				mode = 'Ending'

### Reset prediction 
def resetPrediction():
	global predictedIDset
	global predictedID
	predictedIDset.clear
	predictedID = -1


### Speech
def PicoSpeech(speech):
	call('pico2wave -w ' + ScriptLocation + '/voice.wav \"' + speech + '\" && paplay ' + ScriptLocation + '/voice.wav', shell=True)
	
### Program shutdown sequence
def quit_program(ShutDown):
	print('[INFO] Quit Program')
	cap.release()
	c.close()
	conn.close()
	lcd.lcd_clear()
	lcd.backlight(0)
	GPIO.cleanup()
	if ShutDown:
		call('sudo shutdown -h now', shell=True)
	else:
		quit()

################################### END Function Blocks ###################################

### Assign events to buttons
GPIO.add_event_detect(Button_1, GPIO.BOTH, callback=button1_callback)
GPIO.add_event_detect(Button_2, GPIO.FALLING, callback=button2_callback, bouncetime=300)
GPIO.add_event_detect(Button_3, GPIO.FALLING, callback=button3_callback, bouncetime=300)
GPIO.add_event_detect(Button_4, GPIO.FALLING, callback=button4_callback, bouncetime=300)
GPIO.add_event_detect(Button_5, GPIO.FALLING, callback=button5_callback, bouncetime=300)

### Voice Practice
lcd.lcd_display_string('Voice Practice'.ljust(16), 1, 0)
print('[INFO] Voice Practice...') 
PicoSpeech('hh Audio ready.')

### Start capturing image

cap = cv2.VideoCapture(0)

### Ready to enter main program loop

lcd.lcd_display_string('Program Ready'.ljust(16), 1, 0)
print('[INFO] Program Ready')
sleep(1)
lcd.lcd_clear()

### Main program loop
try:
	while(endProgramCount  <  endProgramAfter):
		sleepTimer = 0
		NewSpeech = ''
		msg1 = format(WeekDay[datetime.now().weekday()]) + datetime.now().strftime(" %d-%b %H:%M")
		msg2 = 'REC          OFF'
		if 'Saving'==mode:
			lcd.lcd_display_string('Saving...'.ljust(16), 1, 0)
			lcd.lcd_display_string('Please wait.'.ljust(16), 2, 0)
		
			sql = '''INSERT INTO people(name, date_created) VALUES (?,?)'''
			c.execute(sql, (name, datetime.now()))
			conn.commit()
			rowid = c.lastrowid
			if newTrainer:
				recogniser.train(trainingImg, np_array([rowid]*len(trainingImg)))
			else:
				recogniser.update(trainingImg, np_array([rowid]*len(trainingImg)))
			print('write facetrainer.yml')
			recogniser.write(ScriptLocation + '/facetrainer.yml')
			newTrainer = False
			resetPrediction()
			trainingImg = []
			name = ''
			alphabet_index = 0
			mode = 'Detecting'
		elif 'Updating'== mode:
			lcd.lcd_display_string('Updating...'.ljust(16), 1, 0)
			lcd.lcd_display_string('Please wait.'.ljust(16), 2, 0)
			recogniser.update(trainingImg, np_array([predictedID]*len(trainingImg)))
			recogniser.write(ScriptLocation + '/facetrainer.yml')
			resetPrediction()
			trainingImg = []
			mode = 'Detecting'
		elif 'Predicting'==mode:
			if len(predictedIDset)==0:
				for Img in trainingImg:
					id, confidence = recogniser.predict(Img)
					predictedIDset.add(id)
			else:
				if not (predictedID in predictedIDset):
					for i in predictedIDset:
						predictedID = i
						sql = '''SELECT name FROM people WHERE ROWID = ? '''
						c.execute(sql, (str(predictedID),))
						predictedName = c.fetchone()[0]
		elif 'Detecting'==mode and not newTrainer:
			for i in range(5):
				ret, frame = cap.read()
			frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
			faces = face_cascade.detectMultiScale(frame, 1.3, 5)
			for (x,y,w,h) in faces:
				msg1 = 'Face Detected!'
				id, confidence = recogniser.predict(frame[y:y+h, x:x+w])
				sql = '''SELECT name FROM people WHERE ROWID = ? '''
				c.execute(sql, (str(id),))
				if confidence < 55:
					if confidence < 20:
						msg1 = 'Hello'
						msg2 = c.fetchone()[0]
					elif confidence < 25:
						msg1 = 'How are you'
						msg2 = c.fetchone()[0] + '?'
					elif confidence < 30:
						msg1 = 'You must be'
						msg2 = c.fetchone()[0]
					elif confidence < 35:
						msg1 = 'How is it going'
						msg2 = c.fetchone()[0] + '?'
					elif confidence < 40:
						msg1 = 'Nice to see you'
						msg2 = c.fetchone()[0]
					elif confidence < 45:
						msg1 = 'I think you are'
						msg2 = c.fetchone()[0]
					elif confidence < 50:
						msg1 = 'You look like'
						msg2 = c.fetchone()[0]
					else:
						msg1 = 'You may be'
						msg2 = c.fetchone()[0] + '?'
					NewSpeech = msg1 + ' ' + msg2 
				elif confidence < 100:
					msg1 = str(round(100-confidence))+ '% similar to'
					msg2 = c.fetchone()[0]
		if 'Recording' == mode:
			newLCDLine1 = indicator + ' Detecting...'
			newLCDLine2 = str(len(trainingImg)) + ' detected'
		elif 'Name' == mode:
			if NameCursorBlink:
				newLCDLine1 = 'Name:'+name
				NameCursorBlink = False
			else:
				newLCDLine1 = 'Name:'+name+alphabet[alphabet_index]
				NameCursorBlink = True
			newLCDLine2 = 'Sve <   >  E   C'
		elif 'Predicting' == mode:
			newLCDLine1 = 'Are you ' + predictedName + '?'
			newLCDLine2 = '   Yes    No'
		elif 'Ending' == mode:
			newLCDLine1 = 'End Program in'
			newLCDLine2 = str(endProgramAfter - endProgramCount)
		else:
			newLCDLine1 = msg1
			newLCDLine2 = msg2
			
		if LCDLine1 != newLCDLine1:
			LCDLine1 = newLCDLine1
			lcd.lcd_display_string(LCDLine1.ljust(16), 1, 0)
		
		if LCDLine2 != newLCDLine2:
			LCDLine2 = newLCDLine2
			lcd.lcd_display_string(LCDLine2.ljust(16), 2, 0)
		if '' != NewSpeech and NewSpeech != OldSpeech:
			PicoSpeech(NewSpeech)
			OldSpeech = NewSpeech
	
### Out of main program loop 
	quit_program(True)
	
### Program interruption		
except KeyboardInterrupt:
	quit_program(False)